@@ -38,7 +38,7 @@ use CGI::FormBuilder;
 use HTML::Entities;
 use URI::Escape;
 
-our $VERSION = '0.99';
+our $VERSION = '1.00';
 
 =encoding utf8
 
@@ -123,8 +123,11 @@ connection.
                 label_column => 'name',
             },
         },
-        custom_columns => {
-            division_news => {
+	table_class => 'table table-bordered',
+	paginate_table_class => 'table table-borderless',
+        custom_columns => [
+            {
+                name => "division_news",
                 raw_column => "division",
                 transform  => sub {
                     my $division_name = shift;
@@ -134,13 +137,13 @@ connection.
                     return "<a href='$search'>$label</a>";
                 },
             },
-        },
+        ],
         auth => {
             view => {
                 require_login => 1,
             },
             edit => {
-                require_role => 1,
+                require_role => 'Admin',
             },
         },
     );
@@ -201,13 +204,19 @@ Specify one or more 'where' clauses to use to filter the table. For example:
         ...
     );
 
-This would cause only rows with an user_id of 1000 to be displayed.
+This would cause only rows with an user_id of 1000 to be displayed in listings
+and search results, viewed, edited etc.
 
 The C<where_filter> parameter takes a hashref describing the WHERE clause, as 
 used by L<Dancer::Plugin::Database>'s C<quick_select> convenience method for
 example - see the 
 L<where clause documentation in Dancer::Plugin::Database::Core::Handle|Dancer::Plugin::Database::Core::Handle/WHERE-clauses-as-hashrefs>.
 
+Alternatively, if the filter condition needs to be calculated at runtime (for
+example, based on the logged in user calling it), then you can provide a coderef
+which returns the WHERE clause hashref - for instance:
+
+  where_filter => sub { { customer_id => logged_in_user()->{customer_id} } },
 
 =item C<db_connection_name> (optional)
 
@@ -410,16 +419,55 @@ C<label_column>.
 
 =item C<custom_columns>
 
-A hashref to specify custom columns to appear in the list view of an entity.
-The keys of the hash are custom column names, the values hashrefs specifying
-a C<raw_column> indicating a column from the table that should be selected to
-build the custom column from,  and C<transform>, a subref to be used as a
+An arrayref of hashrefs to specify custom columns to appear in the list view 
+of an entity.  (Previously, this was just a hashref of column names and specs,
+and this style is still supported for backwards compatibility, but is deprecated
+because it leaves the order of the columns unpredictable.)
+
+The keys of each hash are C<name>, the name to use for this custom column,
+C<raw_column> indicating a column from the table that should be selected to
+build the custom column from, and C<transform>, a subref to be used as a
 HTML::Table::FromDatabase callback on the resulting column.  If no
 C<transform> is provided, sub { return shift; } will be used.
 
-If C<raw_column> consists of anything other than letters, numbers, and underscores,
-it is passed in raw, so you could put something like "NOW()"  or "datetime('now')"
-in there and it should work as expected.
+For a somewhat spurious example:
+
+    ...
+    custom_columns => [
+        {
+            name => 'email_provider',
+            raw_column => 'email',
+            transform => sub {
+                my $value = shift;
+                return (split /@/, 1)[1];
+            },
+        },
+    ],
+    ...
+
+
+The C<transform> code ref is passed to L<HTML::Table::FromDatabase> as a
+callback for that column, so it can do anything a
+L<HTML::Table::FromDatabase callback|HTML::Table::FromDatabase/CALLBACKS>
+can do.  In particular, the coderef will receive the value of the
+column as the first parameter, but also a reference to the whole row hashref
+as the second parameter, so you can do a variety of cunning things.
+
+An example of a custom column whose C<transform> coderef uses the row
+hashref to get other values for the same row could be:
+
+    ...
+    custom_columns => [
+        {
+            name => 'salutation',
+            raw_column => 'name',
+            transform => sub {
+                my ($name_value, $row) = @_;
+                return "Hi, $row->{title} $name_value!";
+            },
+        }
+    ],
+    ...
 
 =item C<auth>
 
@@ -448,6 +496,14 @@ editing), you can use the shorthand:
     },
 
 
+=item C<table_class>
+
+This provides a CSS class for the tables.
+
+=item C<paginate_table_class>
+
+This provides a CSS class for the tables paginate buttons.
+
 =cut
 
 sub simple_crud {
@@ -532,7 +588,7 @@ sub simple_crud {
         get _construct_url(
             $args{dancer_prefix}, $args{prefix}, "/delete/:id"
             ) => sub {
-            return _apply_template(<<CONFIRMDELETE, $args{'template'});
+            return _apply_template(<<CONFIRMDELETE, $args{'template'}, $args{'record_title'});
 <p>
 Do you really wish to delete this record?
 </p>
@@ -552,7 +608,9 @@ CONFIRMDELETE
         my $delete_handler = sub {
             my ($id) = params->{record_id} || splat;
             my $dbh = database($args{db_connection_name});
-            $dbh->quick_delete($table_name, { $key_column => $id })
+            my $where = _get_where_filter_from_args(\%args);
+            $where->{$key_column} = $id;
+            $dbh->quick_delete($table_name, $where)
                 or return _apply_template("<p>Failed to delete!</p>",
                 $args{'template'});
 
@@ -580,8 +638,13 @@ sub _create_view_handler {
 
     my $dbh = database($args->{db_connection_name});
 
-    # a hash containing the current values in the database
-    my $values_from_database = $dbh->quick_select($table_name, { $key_column => $id });
+    # a hash containing the current values in the database.  Take where_filter
+    # into account, so we can't fetch a row if it doesn't match the filter
+    # (otherwise people could load any record they wished just by changing the
+    # ID in the URL, which would be considered a Bad Thing)
+    my $where = _get_where_filter_from_args($args);
+    $where->{$key_column} = $id;
+    my $values_from_database = $dbh->quick_select($table_name, $where);
 
     # Find out about table columns:
     my $all_table_columns = _find_columns($dbh, $args->{db_table});
@@ -615,8 +678,10 @@ sub _create_add_edit_route {
     # a hash containing the current values in the database
     my $values_from_database;
     if ($id) {
+        my $where = _get_where_filter_from_args($args);
+        $where->{$key_column} = $id;
         $values_from_database
-            = $dbh->quick_select($table_name, { $key_column => $id });
+            = $dbh->quick_select($table_name, $where);
     }
 
     # Find out about table columns:
@@ -818,9 +883,12 @@ sub _create_add_edit_route {
         my $success;
         if (exists params('route')->{id}) {
 
-            # We're editing an existing record
-            $success = $dbh->quick_update($table_name,
-                { $key_column => params('route')->{id} }, \%params);
+            # We're editing an existing record - make sure the WHERE clause
+            # hashref incorporates where_filter, if in use, so that users can't
+            # edit stuff they shouldn't be able to
+            my $where = _get_where_filter_from_args($args);
+            $where->{$key_column} = params('route')->{id};
+            $success = $dbh->quick_update($table_name, $where, \%params);
             $verb = 'update';
         } else {
             $success = $dbh->quick_insert($table_name, \%params);
@@ -855,11 +923,11 @@ sub _create_add_edit_route {
             #);
             return _apply_template(
                 "<p>Unable to $verb $args->{record_title}</p>",
-                $args->{'template'});
+                $args->{'template'}, $args->{'record_title'});
         }
 
     } else {
-        return _apply_template($form->render, $args->{'template'});
+        return _apply_template($form->render, $args->{'template'}, $args->{'record_title'});
     }
 }
 
@@ -871,6 +939,9 @@ sub _create_list_handler {
 
     my $display_columns = $args->{'display_columns'};
 
+    my $table_class = $args->{'table_class'} || '';
+    my $paginate_table_class = $args->{'paginate_table_class'} || '';
+
     # If display_columns argument was passed, filter the column list to only
     # have the ones we asked for.
     if (ref $display_columns eq 'ARRAY') {
@@ -890,12 +961,16 @@ sub _create_list_handler {
     my $searchfield_options = join(
         "\n",
         map {
+            my $friendly_name = $_->{COLUMN_NAME};
+            if ($args->{labels}{$_->{COLUMN_NAME}}) {
+                $friendly_name = $args->{labels}{$_->{COLUMN_NAME}};
+            }
             my $sel
                 = (defined params->{searchfield}
                     && params->{searchfield} eq $_->{COLUMN_NAME})
                 ? "selected"
                 : "";
-            "<option $sel value='$_->{COLUMN_NAME}'>$_->{COLUMN_NAME}</option>"
+            "<option $sel value='$_->{COLUMN_NAME}'>$friendly_name</option>"
             } @$columns
     );
     my @searchtypes = (
@@ -972,8 +1047,23 @@ SEARCHFORM
     }
 
     my @custom_cols;
-    foreach my $column_alias ( keys %{ $args->{custom_columns} || {} } ) {
-        my $raw_column = $args->{custom_columns}{$column_alias}{raw_column}
+
+    # For backwards compatibility, understand custom_columns being a hashref,
+    # and translate it
+    if (ref $args->{custom_columns} eq 'HASH') {
+        my @custom_cols_list;
+        for my $column_alias (keys %{ $args->{custom_columns} }) {
+            push @custom_cols_list, {
+                name => $column_alias,
+                %{ $args->{custom_columns}{$column_alias} }
+            };
+        }
+        $args->{custom_columns} = \@custom_cols_list;
+    }
+
+    for my $custom_col_spec (@{ $args->{custom_columns} || [] }) {
+        my $column_alias = $custom_col_spec->{name};
+        my $raw_column = $custom_col_spec->{raw_column}
             or die "you must specify a raw_column that "
                  . "$column_alias will be built using";
         if ($raw_column =~ /^[\w_]+$/) {
@@ -987,10 +1077,10 @@ SEARCHFORM
 
     my $col_list = join(
         ',',
-        map({ $table_name . "." . $dbh->quote_identifier($_) . " AS " .
-                  $dbh->quote_identifier($args->{labels}{$_} || $_)
-            }
-            @select_cols),
+        map(
+            { $table_name . "." . $dbh->quote_identifier($_) }
+            @select_cols
+        ),
         @foreign_cols,    # already assembled from quoted identifiers
         @custom_cols,
     );
@@ -1023,9 +1113,14 @@ SEARCHFORM
     }
 
     # If we have a query or a where_filter, we need to assemble a WHERE clause...
-    my $where_filter = $args->{where_filter};
+    my $where_filter = _get_where_filter_from_args($args);
     if (length $q || $where_filter) {
-        my ($where_filter_sql, @where_filter_binds) = $dbh->generate_where_clauses($where_filter);
+        
+        # Turn the $where_filter hashref into some SQL clauses and bind params,
+        # which we'll add to with the user's search params shortly
+        my ($where_filter_sql, @where_filter_binds)
+            = $dbh->generate_where_clauses($where_filter);
+
         my (@search_wheres, @search_binds);
         if (length $q) {    # this nested code is all for queries in $q
             my ($column_data)
@@ -1088,8 +1183,6 @@ SEARCHFORM
             . "<p>";
     }
 
-    ## Build a hash to add sorting CGI parameters + URL to each column header.
-    ## (will be used with HTML::Table::FromDatabase's "-rename_columns" parameter.
     my %columns_sort_options;
     if ($args->{sortable}) {
         my $qt              = uri_escape($q);
@@ -1109,20 +1202,43 @@ SEARCHFORM
         my $opposite_order_by_direction
             = ($order_by_direction eq "asc") ? "desc" : "asc";
 
+        # Get a list of all columns (normal, and custom_columns), then assemble
+        # the names and labels to pass to HTML::Table::FromDatabase
+        my @all_cols = map { $_->{COLUMN_NAME} } @$columns;
         %columns_sort_options = map {
-            my $col_name       = $_->{COLUMN_NAME};
-            my $col = $args->{labels}{$col_name} || $col_name;
+            my $col_name       = $_;
             my $direction      = $order_by_direction;
             my $direction_char = "";
+            my $friendly_name  = $col_name;
+            if ($args->{labels}{$col_name}) {
+                $friendly_name = $args->{labels}{$col_name};
+            } else {
+                $friendly_name = _prettify_column_name($friendly_name);
+            }
             if ($col_name eq $order_by_column) {
                 $direction = $opposite_order_by_direction;
                 $direction_char = ($direction eq "asc") ? "&uarr;" : "&darr;";
             }
             my $url = _external_url($args->{dancer_prefix}, $args->{prefix})
                 . "?o=$col_name&d=$direction&q=$q&searchfield=$sf&searchtype=$st";
-            $col =>
-                "<a href=\"$url\">$col&nbsp;$direction_char</a>";
-        } @$columns;
+            $col_name =>
+                "<a href=\"$url\">$friendly_name&nbsp;$direction_char</a>";
+        } @all_cols;
+
+        # And for custom columns, do the prettification, but don't include a
+        # link for sorting - as we can't sort by them currently (the sorting is
+        # done by SQL, and the custom column values are calculated after we get
+        # the results from the SQL query, so to support sorting by them we'd
+        # have to stop getting the database to sort the data and sort it
+        # ourselves afterwards).
+        if (exists $args->{custom_columns}) {
+            for my $custom_column_name (
+                map { $_->{name} } @{ $args->{custom_columns} }
+            ) {
+                $columns_sort_options{$custom_column_name}
+                    = _prettify_column_name($custom_column_name);
+            }
+        }
 
         if (exists $args->{foreign_keys} and exists $args->{foreign_keys}{$order_by_column}) {
                 my $fk = $args->{foreign_keys}{$order_by_column};
@@ -1131,7 +1247,8 @@ SEARCHFORM
         }
 
         $query .= " ORDER BY "
-            . $dbh->quote_identifier($order_by_table) . "." . $dbh->quote_identifier($order_by_column)
+            . $dbh->quote_identifier($order_by_table) . "."
+            . $dbh->quote_identifier($order_by_column)
             . " $order_by_direction ";
     }
 
@@ -1152,26 +1269,26 @@ SEARCHFORM
         my $url = _external_url($args->{dancer_prefix}, $args->{prefix})
             . "?o=$o&d=$d&q=$qt&searchfield=$sf&searchtype=$st";
         $html .= "<p>";
+	$html .= "<table class=\"$paginate_table_class\"><tr>";
+
         if ($page > 0) {
             $html
                 .= sprintf(
-                "<a href=\"%s&p=%d\">&larr;&nbsp;prev.&nbsp;page</a>",
+                "<td><a href=\"%s&p=%d\">&larr;&nbsp;prev.&nbsp;page</a></td>",
                 $url, $page - 1)
         } else {
-            $html .= "&larr;&nbsp;prev.&nbsp;page&nbsp";
+            $html .= "<td>&larr;&nbsp;prev.&nbsp;page&nbsp</td>";
         }
-        $html .= "&nbsp;" x 5;
         $html .= sprintf(
-            "Showing page %d (records %d to %d)",
+            "<td>Showing page %d (records %d to %d)",
             $page + 1,
             $offset + 1,
             $offset + 1 + $limit
         );
-        $html .= "&nbsp;" x 5;
-        $html
-            .= sprintf("<a href=\"%s&p=%d\">next&nbsp;page&nbsp;&rarr;</a>",
+        $html .= "</td>";
+        $html .= sprintf("<td><a href=\"%s&p=%d\">next&nbsp;page&nbsp;&rarr;</a>",
             $url, $page + 1);
-        $html .= "<p>";
+        $html .= "</td></tr></table>";
 
         $query .= " LIMIT $limit OFFSET $offset ";
     }
@@ -1189,12 +1306,14 @@ SEARCHFORM
     }
 
     my @custom_callbacks = ();
-    foreach my $column_alias ( keys %{ $args->{custom_columns} || {} } ) {
-        push @custom_callbacks, { 
-            column=>$column_alias, 
-            transform=> ($args->{custom_columns}->{$column_alias}->{transform} or sub { return shift;}),
+    for my $custom_col_spec (@{ $args->{custom_columns} || [] } ) {
+        push @custom_callbacks, {
+            column=>$custom_col_spec->{name}, 
+            transform=> ($custom_col_spec->{transform} or sub { return shift;}),
         };
     }
+
+
     my $table = HTML::Table::FromDatabase->new(
         -sth       => $sth,
         -border    => 1,
@@ -1229,8 +1348,8 @@ SEARCHFORM
             @custom_callbacks,
         ],
         -rename_headers      => \%columns_sort_options,
-        -auto_pretty_headers => 1,
         -html                => 'escape',
+        -class               => "$table_class",
     );
 
     $html .= $table->getTable || '';
@@ -1262,14 +1381,14 @@ function delrec(record_id) {
 DELETEJS
     }
 
-    return _apply_template($html, $args->{'template'});
+    return _apply_template($html, $args->{'template'}, $args->{'record_title'});
 }
 
 sub _apply_template {
-    my ($html, $template) = @_;
+    my ($html, $template, $title) = @_;
 
     if ($template) {
-        return template $template, { simple_crud => $html };
+        return template $template, { simple_crud => $html, record_title => $title };
     } else {
         return engine('template')->apply_layout($html);
     }
@@ -1468,6 +1587,39 @@ sub _defined_or_empty {
     return defined($v) ? $v : "";
 }
 
+# where_filter  "if it's a coderef, call it and check it gave us a hashref to
+# use, otherwise expect it to be a hashref" logic as we need this in several
+# places.  Returns a hashref, ready for us to add other stuff to in most cases
+# (e.g. usually we'd call this, then add $key_column => ... to it)
+sub _get_where_filter_from_args {
+    my $args = shift;
+    return unless $args->{where_filter};
+
+    if (ref $args->{where_filter} eq 'HASH') {
+        return $args->{where_filter};
+    } elsif (ref $args->{where_filter} eq 'CODE') {
+        my $result = $args->{where_filter}->();
+        if (ref $result eq 'HASH') {
+            return $result;
+        } else {
+            # TODO: better error reporting, so we know which one caused
+            # the problem
+            die "where_filter coderef didn't return a hashref!";
+        }
+    } else {
+        die "Invalid where_filter";
+    }
+}
+
+sub _prettify_column_name {
+    my $name = shift;
+    for ($name) {
+        $_ = lc;
+        s{_}{ }g;
+        s{\b(\w)}{\u$1}g;
+    }
+    return $name;
+}
 
 =back
 
@@ -1491,6 +1643,9 @@ C<simple_crud>, in which case what you say goes.)
 
 =head1 Hooks
 
+Hooks are provided, which can be used in the normal Dancer way, using the 
+C<hook> keyword.
+
 =head2 add_edit_row (deprecated, use add_edit_row_pre_save)
 
 You can use the same code from your add_edit_row hook in an add_edit_row_pre_save
@@ -1513,7 +1668,15 @@ C<dbh> giving you the instance of the handle used to save the entity
 'create new' or 'update').
 
 For instance, if you were dealing with a users table, you could use the
-pre_save hook to hash the password before storing it.
+pre_save hook to hash the password before storing it - assuming for the sake
+of example that you have a C<hash_pw()> function to return a hashed password:
+
+  hook add_edit_row_pre_save => sub {
+      my $args = shift;
+      if ($args->{table_name} eq 'user') {
+          $args->{params}{password} = hash_pw($args->{params}{password});
+      }
+  };
 
 =head1 AUTHOR
 
